/* This file is a part of the Amalthea library.
 *
 * Copyright (C) 2018-2021 Eugene 'Vindex' Stulin
 *
 * The Amalthea library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This libary is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

module amalthea.dataprocessing;

public import amalthea.libcore;

import std.algorithm, std.array, std.json;


/*******************************************************************************
 * Implementation of the intersection operation for arrays.
 */
T calcIntersection(T)(T[] arrays ...) {
    import std.algorithm.setops : setIntersection;
    if (arrays.length == 1) return arrays[0].dup;
    T result = setIntersection(sort(arrays[0]), sort(arrays[1])).array;
    for (auto i = 2; i < arrays.length; ++i) {
        result = setIntersection(result, sort(arrays[i])).array;
    }
    return result;
}
///
unittest {
    int[] arr = calcIntersection([3, 4, 1, -15], [9, 3, 1], [1, 99, 3]);
    assert(arr == [1, 3]);
}
deprecated alias opIntersection = calcIntersection;


/*******************************************************************************
 * Implementation of the union operation for arrays.
 */
T calcUnion(T)(T[] arrays ...) {
    T arrays0 = arrays[0];
    if (arrays.length == 1) return arrays0;
    T result = arrays[0];
    for (size_t i = 1; i < arrays.length; ++i) {
        foreach(el; arrays[i]) {
            if (!canFind(result, el)) result ~= el;
        }
    }
    return std.algorithm.sort(result).array;
}
///
unittest {
    int[] arr = calcUnion([3, 4, 1, -15], [9, 3, 1], [1, 99, 3]);
    assert(arr == [-15, 1, 3, 4, 9, 99]);
}
deprecated alias opOnion = calcUnion;


/*******************************************************************************
 * Implementation of the complement operation for arrays.
 */
T calcComplement(T)(T[] arrays ...) {
    if (arrays.length == 1) return arrays[0].dup;
    T result = array(setDifference(sort(arrays[0]), sort(arrays[1])));
    for (auto i = 2; i < arrays.length; ++i) {
        result = array(setDifference(result, sort(arrays[i])));
    }
    return result;
}
///
unittest {
    int[] arr = calcComplement([3, 4, 1, -15], [9, 3, 1]);
    assert(arr == [-15, 4]);
}
deprecated alias opComplement = calcComplement;


/*******************************************************************************
 * The function removes consecutive string duplicates.
 */
string removeDuplicateConsecutiveSubstring(string s, string substring) {
    while(s.canFind(substring~substring)) {
        s = s.replace(substring~substring, substring);
    }
    return s;
}
///
unittest {
    string s = removeDuplicateConsecutiveSubstring("AAAAbooeAAAuuvAww", "A");
    assert(s == "AbooeAuuvAww");
    s = removeDuplicateConsecutiveSubstring("AAAAbcooeAAAgguvAww", "AA");
    assert(s == "AAbcooeAAAgguvAww");
}


/*******************************************************************************
 * The function removes duplicate elements from an array.
 */
ref T[] removeDuplicate(T)(ref T[] arr) {
    if (arr.empty) return arr;
    for (auto i = arr.length-1; i > 0; --i) {
        if (arr[0 .. i].canFind(arr[i])) {
            arr = arr.remove(i);
        }
    }
    return arr;
}
///
unittest {
    int[] values = [1, 3, 1, 7, 2, 19, 5, 2, 2];
    removeDuplicate(values);
    assert(values == [1, 3, 7, 2, 19, 5]);
}


/*******************************************************************************
 * The function arranges the associative array
 * by the required sorting method (string "by" is equal to "values" or "keys"),
 * returns an array of keys and an array of values.
 */
void orderAssociativeArray(alias by, V, K)(in V[K] aarray,
                                               out K[] orderedKeys,
                                               out V[] orderedValues)
if (by == "values" || by == "keys") {
    if (aarray.length == 0) {
        return;
    }
    if (by == "keys") {
        orderedKeys = aarray.keys.dup;
        orderedKeys.sort;
        orderedValues.length = orderedKeys.length;
        foreach(i, k; orderedKeys) {
            orderedValues[i] = aarray[k];
        }
    } else {
        orderedValues = aarray.values.dup;
        orderedValues.sort;
        orderedKeys.length = orderedValues.length;
        size_t[] passedIndexes;
        foreach(k; aarray.keys) {
            foreach(i, v; orderedValues) {
                if (passedIndexes.canFind(i)) {
                    continue;
                }
                if (aarray[k] == v) {
                    orderedKeys[i] = k;
                    passedIndexes ~= i;
                    break;
                }
            }
        }
    }
}


/*******************************************************************************
 * The function arranges the associative array by keys,
 * returns an array of keys and an array of values.
 */
void orderAssociativeArrayByKeys(V, K)(in V[K] aarray,
                                       out K[] orderedKeys,
                                       out V[] orderedValues) {
    orderAssociativeArray!"keys"(aarray, orderedKeys, orderedValues);
}


/*******************************************************************************
 * The function arranges the associative array by values,
 * returns an array of keys and an array of values.
 */
void orderAssociativeArrayByValues(V, K)(in V[K] aarray,
                                         out K[] orderedKeys,
                                         out V[] orderedValues) {
    orderAssociativeArray!"values"(aarray, orderedKeys, orderedValues);
}


deprecated alias orderMap = orderAssociativeArray;
deprecated alias orderMapByKeys = orderAssociativeArrayByKeys;
deprecated alias orderMapByValues = orderAssociativeArrayByValues;


unittest {
    string[string] aarray = ["Canada" : "Ottawa",
                             "Belarus": "Minsk",
                             "USA"    : "Washington",
                             "Myanmar": "Naypyidaw",
                             "France" : "Paris"];
    string[] countries;
    string[] capitals;
    orderAssociativeArrayByKeys(aarray, countries, capitals);
    assert(countries == ["Belarus","Canada","France","Myanmar",  "USA"]);
    assert(capitals  == ["Minsk",  "Ottawa","Paris", "Naypyidaw","Washington"]);
    orderAssociativeArrayByValues(aarray, countries, capitals);
    assert(countries == ["Belarus","Myanmar",  "Canada","France","USA"]);
    assert(capitals  == ["Minsk",  "Naypyidaw","Ottawa","Paris", "Washington"]);

    auto aarray2 = ["hundred": 100, "much": 100, "one": 1, "two": 2, ];
    string[] keys;
    int[] values;
    orderAssociativeArrayByValues(aarray2, keys, values);
    assert(keys == ["one", "two", "hundred", "much"]
        || keys == ["one", "two", "much", "hundred"]);
    assert(values == [1, 2, 100, 100]);
}


/*******************************************************************************
 * This function combines two associative arrays, returns new array.
 */
V[K] mixAssociativeArrays(K,V)(V[K] firstArray, V[K] secondArray) {
    V[K] array = firstArray.dup;
    foreach(k, v; secondArray) {
        if (k !in array) {
            array[k] = v;
        }
    }
    return array;
}
deprecated alias mixMaps = mixAssociativeArrays;
deprecated alias combineMaps = mixAssociativeArrays;
deprecated alias combineAssociativeArrays = mixAssociativeArrays;


/*******************************************************************************
 * Function to get a value by parameter.
 * Used to process command line arguments.
 * Params:
 *     args = Command line arguments.
 *     flag = Parameter with "--".
 *     altFlag = Synonym for the flat. By default, empty.
 * Returns: The value of the required parameter.
 */
string getOptionValue(string[] args, string flag, string altFlag = "") {
    auto tmpArgs = args.dup;
    return extractValueForFlag(tmpArgs, flag, altFlag);
}
alias getValueForFlag = getOptionValue;


/*******************************************************************************
 * Function for extracting of value by flag.
 * Unlike getOptionValue(), this function changes the passed `args`,
 * removing the flag with a value.
 * Params:
 *     args = Command line arguments. Changes while the function is running.
 *     flag = Parameter with "--".
 *     altFlag = Synonym for the flat. By default, empty.
 * Returns: The value of the required parameter.
 */
string extractOptionValue(ref string[] args, string flag, string altFlag = "") {
    string value;
    ssize_t index1 = -1;

    cycle1: foreach(option; [flag, altFlag]) {
        foreach(i, arg; args) {
            if (arg == option || arg.startsWith(option~"=")) {
                index1 = i;
                flag = option;
                break cycle1;
            }
        }
    }
    if (index1 == -1) return "";
    
    if (args[index1].startsWith(flag~"=")) {
        value = args[index1][(flag~"=").length .. $];
        auto tempArgs = args[0 .. index1];
        if (args.length > index1+1) tempArgs ~= args[index1+1 .. $];
        args = tempArgs;
    } else {
        if (args.length < index1+2) return "";
        value = args[index1+1];
        auto tempArgs = args[0 .. index1];
        if (args.length > index1+2) tempArgs ~= args[index1+2 .. $];
        args = tempArgs;
    }
    return value;
}
alias extractValueForFlag = extractOptionValue;

///
unittest {
    string[] args = "app -a value1 value2 -b parameter --nothing".split;
    assert("parameter" == args.getOptionValue("-b"));
    assert("value1" == args.getOptionValue("-a"));
    assert("" == args.getOptionValue("--nothing"));
    
    assert(args == "app -a value1 value2 -b parameter --nothing".split);

    assert("parameter" == args.extractValueForFlag("-b"));
    assert("app -a value1 value2 --nothing".split == args);
    assert("value1" == args.extractValueForFlag("-a"));
    assert("app value2 --nothing".split == args);

    args = "-a value1 value2 -b parameter --nothing".split;
    assert("parameter" == args.getOptionValue("-b"));
    assert("value1" == args.getOptionValue("-a"));
    assert("" == args.getOptionValue("--nothing"));
    assert("parameter" == args.extractValueForFlag("-b"));
    assert("-a value1 value2 --nothing".split == args);
    assert("value1" == args.extractValueForFlag("-a"));
    assert("value2 --nothing".split == args);

    args = `build --file=/home/user/Project_1/Makefile`.split;
    assert("/home/user/Project_1/Makefile" == args.getOptionValue("--file"));

}


/*******************************************************************************
 * Function to get values as array by flag.
 * Params:
 *     args = Command line arguments.
 *     flag = Parameter with "--".
 *     altFlag = Synonym for the flat. By default, empty.
 * Returns: array of values after the flag.
 */
string[] getOptionRange(string[] args, string flag, string altFlag = "") {
    auto argsCopy = args.dup;
    return extractOptionRange(argsCopy, flag, altFlag);
}
deprecated alias getRangeForFlag = getOptionRange;


/*******************************************************************************
 * Function for extracting of value as array by flag.
 * Unlike getOptionRange(), this function changes the passed arguments (args),
 * removing the flag with a value.
 * Params:
 *     args = Command line arguments. Changes while the function is running.
 *     flag = Parameter with "--".
 *     altFlag = Synonym for the flat. By default, empty.
 * Returns: array of values after the flag.
 */
string[] extractOptionRange(ref string[] args,
                            string flag,
                            string altFlag = "") {
    string[] range;
    ssize_t index1 = args.countUntil(flag);
    ssize_t index2 = -1;

    if (index1 == -1) index1 = args.countUntil(altFlag);
    if (index1 == -1) return range;
    if (index1 != args.length-1 && !args[index1+1].startsWith("-")) {
        foreach(i, arg; args[index1+1 .. $]) {
            if (arg.startsWith("-")) {
                index2 = index1 + i;
                break;
            }
        }
        if (index2 != -1) {
            range = args[index1+1 .. index2+1];
            args = args[0 .. index1] ~ args[index2+1 .. $];
        } else {
            index2 = args.length - 1;
            range = args[index1+1 .. index2+1];
            args = args[0 .. index1];
        }
    }
    return range;
}
deprecated alias extractRangeForFlag = extractOptionRange;


///
unittest {
    string[] args = "app -a value1 value2 -b parameter --nothing".split;
    assert(["value1", "value2"] == getOptionRange(args, "-a"));
    assert(["parameter"] == getOptionRange(args, "-b"));
    string[] nothing = getOptionRange(args, "--nothing");
    assert(nothing.empty);

    args = "app --alt value1 value2 -b parameter --nothing".split;
    assert(["value1", "value2"] == getOptionRange(args, "-a", "--alt"));

    args = "app -a -b thing".split;
    string[] vacuum = getOptionRange(args, "-a");
    assert(vacuum.empty);

    args = "app -a value1 value2 -b parameter --nothing".split;
    assert("value1 value2".split == extractOptionRange(args, "-a"));
    assert("app -b parameter --nothing".split == args);

    args = "-a value1 value2 -b parameter --nothing".split;
    assert("value1 value2".split == extractOptionRange(args, "-a"));
    assert("-b parameter --nothing".split == args);
}


/*******************************************************************************
 * The function converts integer number into an octal form, returns as a string.
 */
string getOctalForm(ulong value) {
    import std.format : formattedWrite;
    auto writer = std.array.appender!string();
    formattedWrite(writer, "%o", value);
    string octalForm = writer.data;
    return octalForm;
}
alias toOctalString = getOctalForm;
///
unittest {
    short eight = 8;
    assert(getOctalForm(eight) == "10");
    assert(33261.getOctalForm == "100755");
}


/*******************************************************************************
 * The function returns the index of the array element by value.
 * Params:
 *     haystack = The array in which the search is performed.
 *     needle = Element value in the array.
 * Returns: The index of the array element. -1 if the element was not found.
 */
ssize_t getIndex(T)(T[] haystack, T needle) {
    foreach(i, el; haystack) {
        if (el == needle) {
            return i;
        }
    }
    return -1;
}


/*******************************************************************************
 * The function returns an array without unnecessary elements.
 */
T[] removeElementsByContent(T)(T[] haystack, T needle) {
    for(ssize_t i; i < haystack.length; ++i) {
        if (haystack[i] == needle) {
            if (haystack.length-1 != i) {
                haystack = haystack[0 .. i] ~ haystack[i+1 .. $];
            } else {
                haystack = haystack[0 .. i];
            }
            --i;
        }
    }
    return haystack.dup;
}
///
unittest {
    auto arr = [1, 8, 17, 4, 23, 8, 8, 3, 13, 19, 8];
    auto res = arr.removeElementsByContent(8);
    assert(res == [1, 17, 4, 23, 3, 13, 19]);

    auto lines = ["", "straw", "", "", "one more straw", "", ""];
    auto straws = lines.removeElementsByContent("");
    assert(straws == ["straw", "one more straw"]);
}


/*******************************************************************************
 * Encode associative array using www-form-urlencoding (copied from std.uri).
 * Params:
 *     values = An associative array containing the values to be encoded.
 * Returns: A string encoded using www-form-urlencoding.
 */
string urlEncode(in string[string] values) {
    import std.uri : encodeComponent;
    import std.array : Appender;
    import std.format : formattedWrite;
    if (values.length == 0) return "";
    Appender!string enc;
    enc.reserve(values.length * 128);
    bool first = true;
    foreach (k, v; values) {
        if (!first) enc.put('&');
        formattedWrite(enc, "%s=%s", encodeComponent(k), encodeComponent(v));
        first = false;
    }
    return enc.data;
}


/*******************************************************************************
 * The function converts JSONValue to double.
 */
double convJSONValueToDouble(std.json.JSONValue json) {
    double number;
    static if (__traits(compiles, JSONType.float_)) {
        switch(json.type) {
            case JSONType.float_:   number = json.floating;            break;
            case JSONType.integer:  number = to!double(json.integer);  break;
            case JSONType.uinteger: number = to!double(json.uinteger); break;
            case JSONType.string:   number = to!double(json.str);      break;
            default: break;
        }
    } else {
        switch(json.type) {
            case JSON_TYPE.FLOAT:    number = json.floating;            break;
            case JSON_TYPE.INTEGER:  number = to!double(json.integer);  break;
            case JSON_TYPE.UINTEGER: number = to!double(json.uinteger); break;
            case JSON_TYPE.STRING:   number = to!double(json.str);      break;
            default: break;
        }
    }
    return number;
}


/*******************************************************************************
 * The function similar to functions "canFind" and "among".
 */
bool isAmong(alias pred = (a, b) => a == b, Value, Values...)
            (Value value, Values values)
if (
    Values.length != 0
) {
    foreach (ref v; values) {
        import std.functional : binaryFun;
        if (binaryFun!pred(value, v)) return true;
    }
    return false;
}
/// ditto
bool isAmong(alias pred = (a, b) => a == b, T)
            (T value, T[] values) {
    foreach (ref v; values) {
        import std.functional : binaryFun;
        if (binaryFun!pred(value, v)) return true;
    }
    return false;
}


/*******************************************************************************
 * Creates new array with specified content.
 */
T[] makeFilledArray(T)(size_t len, T filler) {
    T[] arr = new T[len];
    std.algorithm.fill(arr, filler);
    return arr;
}
///
unittest {
    assert(makeFilledArray(4, 5) == [5, 5, 5, 5]);
    assert(makeFilledArray(10, ' ') == "          ");
}


/*******************************************************************************
 * The function returns true if string is unsigned integer.
 */
deprecated bool stringIsUint(string value) {
    import std.uni : isNumber;
    foreach(c; value) {
        if (!(isNumber(to!dchar(c)))) {
            return false;
        }
    }
    return true;
}
deprecated unittest {
    assert(true  == stringIsUint("11"));
    assert(false == stringIsUint("53.18"));
    assert(false == stringIsUint("-5"));
}
