/* This file is a part of the Amalthea library.
 *
 * Copyright (C) 2018-2021 Eugene 'Vindex' Stulin
 *
 * The Amalthea library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This libary is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

module amalthea.sys;

import core.stdc.stdlib;
import core.stdc.time;
import core.sys.posix.pwd;
import core.sys.posix.grp;
import core.thread : Thread;

public import amalthea.libcore;
import amalthea.fs;
import amalthea.langlocal;
import amalthea.terminal;

alias isTTY = amalthea.terminal.isTTY;
alias isTerminal = amalthea.terminal.isTerminal;

import std.array,
       std.datetime,
       std.datetime.systime,
       std.path,
       std.process,
       std.range,
       std.string;


/*******************************************************************************
 * Common function for calling external programs.
 */
int c_system(string cmd) {
    return core.stdc.stdlib.system(cmd.toStringz);
}


/*******************************************************************************
 * The function returns output of external command.
 */
deprecated string getAppOutput(string command) {
    import std.process : executeShell;
    auto answer = executeShell(command);
    return (0 != answer.status) ? "" : answer.output.strip;
}


/*******************************************************************************
 * The function returns number of seconds since 1970-01-01.
 * Params:
 *     timeString = String with the format 'YYYY-MM-DD hh:mm:ss'
 *                  or 'YYYY-MM-DDThh:mm:ss'.
 *                  If the string is empty, the function returns current time.
 *                  By default, empty.
 */
long unixTime(string timeString = "") {
    if (timeString.empty) {
        return core.stdc.time.time(null);
    }
    timeString = timeString.replace(" ", "T");
    return SysTime(DateTime.fromISOExtString(timeString)).toUnixTime();
}


/*******************************************************************************
 * Gets time string from the std.datetime.systime.SysTime object.
 * Returns: String with format "YYYY-MM-DD HH:MM:SS".
 */
string getTimeString(SysTime sysTime) {
    return sysTime.toISOExtString().replace("T"," ").split('.')[0];
}


/*******************************************************************************
 * Gets time string with current time.
 * Returns: String with format "YYYY-MM-DD HH:MM:SS".
 */
string getTimeString() {
    return getTimeString(Clock.currTime);
}


/*******************************************************************************
 * Gets time string by UNIX time (number of seconds).
 * Returns: String with format "YYYY-MM-DD HH:MM:SS".
 */
string getTimeString(long secondsByUnixTime) {
    return getTimeString(SysTime.fromUnixTime(secondsByUnixTime));
}


/*******************************************************************************
 * The function makes a delay, expressed in seconds.
 */
void sleep(double seconds) {
    ulong microseconds = (seconds <= 0) ? 0 : to!ulong(seconds * (10^^6));
    Thread.sleep(dur!("usecs")(microseconds));
}


/*******************************************************************************
 * The function makes a delay, expressed in milliseconds.
 */
void msleep(ulong milliseconds) {
    Thread.sleep(dur!("msecs")(milliseconds));
}


/*******************************************************************************
 * The function makes a delay, expressed in microseconds.
 */
void usleep(ulong microseconds) {
    Thread.sleep(dur!("usecs")(microseconds));
}


/*******************************************************************************
 * Returns file path to the application (searches among system directories).
 */
string getAppPath(string app) {
    string[] appsRootPaths = environment["PATH"].split(":");
    string appPath;
    foreach(p; appsRootPaths) {
        appPath = p ~ "/" ~ app;
        if (exists(appPath)) {
            return appPath;
        }
    }
    return "";
}


extern(C) private int mkfifo(const char* pathname, uint mode);

/*******************************************************************************
 * The function creates named pipe.
 */
void createPipe(in string pipepath) {
    if (pipepath.exists) {
        enforce(FileStat(pipepath).type == FileType.pipe,
                new FileException("File exists and is not a named pipe."));
        remove(pipepath);
    }
    mkfifo(pipepath.toStringz, 438); //438₁₀ = 0666₈
}


/*******************************************************************************
 * Helper function for getting the standard path to the share directory
 * of the current program.
 */
string getAppShareDir() {
    string currentApp = readLink("/proc/self/exe");
    string shareDir = buildNormalizedPath(dirName(currentApp), "..", "share");
    return shareDir;
}


/*******************************************************************************
 * Helper function for getting the standard path to the help directory
 * of the current program.
 */
string getAppHelpDir() {
    string shareDir = getAppShareDir();
    string helpDir = buildNormalizedPath(shareDir, "help");
    return helpDir;
}


/*******************************************************************************
 * This function returns help text of the current program using
 * standard paths for help file.
 */
string readAppHelpText(string app, string lang = "") {
    if (lang == "") {
        lang = amalthea.langlocal.getSystemLanguage();
    }
    immutable helpDir = getAppHelpDir();
    string defaultHelpFile = buildPath(helpDir, lang, app, "help.txt");
    string englishHelpFile = buildPath(helpDir, "en_US", app, "help.txt");
    if (exists(defaultHelpFile)) {
        return readText(defaultHelpFile);
    } else if (exists(englishHelpFile)) {
        return readText(englishHelpFile);
    }
    throw new FileException("Help file doesn't exist.");
}


/*******************************************************************************
 * The function returns user name by user ID.
 */
string getNameByUID(uint uid) {
    auto p = getpwuid(uid);
    if (p is null) {
        return uid.to!string;
    }
    return (*p).pw_name.fromStringz.to!string;
}


/*******************************************************************************
 * The function returns group name by group ID.
 */
string getNameByGID(uint gid) {
    auto p = getgrgid(gid);
    if (p is null) {
        return gid.to!string;
    }
    return (*p).gr_name.fromStringz.to!string;
}

